// Copyright 2022, FOSS-VG Developers and Contributers
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, version 3.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// version 3 along with this program.
// If not, see https://www.gnu.org/licenses/agpl-3.0.en.html

#include <string>
#include <iostream>
#include <vector>
#include "assert.h++"
#include "../lib/error.h++"

#include "../lib/cli.h++"

int main(int argc, char* argv[]) {
    std::cout << "################################################################################" << std::endl;
    std::cout << "CLI argument parser tests" << std::endl;
    std::cout << "################################################################################" << std::endl;

    // Valid parameter test ############################################
    // This test exercises the basic functionality of the arguments parser
    // with many variations of valid input on a single command line.
    //
    // simulated command line:
    // test -0 -ab -12345 --long-flag -cconcatenated -d separate-value -efdouble-concatenated "positional argument 0" -gh concatenated-separate-value --long-argument-with-value-included="included value" --long-argument-with-value-separated "separate value" "positional argument 1" "positional argument 2"

    std::vector<CLI::Flag> validTestFlags;
    validTestFlags.push_back(CLI::Flag('0', "00000", "a short flag on its own"));
    validTestFlags.push_back(CLI::Flag('a', "aaaaa", "concatenated short flags"));
    validTestFlags.push_back(CLI::Flag('b', "bbbbb", "concatenated short flags"));
    validTestFlags.push_back(CLI::Flag('1', "11111", "more than two concatenated short flags"));
    validTestFlags.push_back(CLI::Flag('2', "22222", "more than two concatenated short flags"));
    validTestFlags.push_back(CLI::Flag('3', "33333", "more than two concatenated short flags"));
    validTestFlags.push_back(CLI::Flag('4', "44444", "more than two concatenated short flags"));
    validTestFlags.push_back(CLI::Flag('5', "55555", "more than two concatenated short flags"));
    validTestFlags.push_back(CLI::Flag('l', "long-flag", "long flags"));
    validTestFlags.push_back(CLI::Flag('e', "eeeee", "short flags concatenated with an argument that is concatenated with its value"));
    validTestFlags.push_back(CLI::Flag('g', "ggggg", "short flags concatenated with an argument that has a separate value"));
    validTestFlags.push_back(CLI::Flag('6', "66666", "unused flag"));

    std::vector<CLI::UnpositionalArgument> validTestUnpositionalArguments;
    validTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('c', "ccccc", "VALUE", "short argument concatenated with its value"));
    validTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('d', "ddddd", " VALUE", "short argument with separate value"));
    validTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('f', "fffff", "VALUE", "short argument concatenated with a flag and its value"));
    validTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('h', "hhhhh", " VALUE", "short argument concatenated with a flag with separate value"));
    validTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('x', "long-argument-with-value-included", "VALUE", "long argument with its value included using ="));
    validTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('y', "long-argument-with-value-separated", " VALUE", "long argument with separate value"));
    validTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('z', "zzzzz", "NOPE", "unused argument"));

    std::vector<CLI::PositionalArgument> validTestPositionalArguments;
    validTestPositionalArguments.push_back(CLI::PositionalArgument("argument0", "positional argument between optional parameters"));
    validTestPositionalArguments.push_back(CLI::PositionalArgument("argument1", "positional argument"));
    validTestPositionalArguments.push_back(CLI::PositionalArgument("argument2", "positional argument"));

    const char** validTestParameterList = new const char*[17];
    validTestParameterList[ 0] = "test";
    validTestParameterList[ 1] = "-0";
    validTestParameterList[ 2] = "-ab";
    validTestParameterList[ 3] = "-12345";
    validTestParameterList[ 4] = "--long-flag";
    validTestParameterList[ 5] = "-cconcatenated";
    validTestParameterList[ 6] = "-d";
    validTestParameterList[ 7] = "separate-value";
    validTestParameterList[ 8] = "-efdouble-concatenated";
    validTestParameterList[ 9] = "positional argument 0";
    validTestParameterList[10] = "-gh";
    validTestParameterList[11] = "concatenated-separate-value";
    validTestParameterList[12] = "--long-argument-with-value-included=included value";
    validTestParameterList[13] = "--long-argument-with-value-separated";
    validTestParameterList[14] = "separate value";
    validTestParameterList[15] = "positional argument 1";
    validTestParameterList[16] = "positional argument 2";
    int validTestParameterCount = 17;

    CLI::ArgumentsParser validTestParameterParser = CLI::ArgumentsParser(validTestParameterCount, validTestParameterList, validTestFlags, validTestUnpositionalArguments, validTestPositionalArguments);

    ASSERT(!validTestParameterParser.wrongUsage);
    ASSERT(validTestParameterParser.programName == std::string("test"));
    ASSERT(!validTestParameterParser.getFlag('0').isError);
    ASSERT(validTestParameterParser.getFlag('0').value);
    ASSERT(validTestParameterParser.getFlag(std::string("00000")).value);
    ASSERT(!validTestParameterParser.getFlag('a').isError);
    ASSERT(validTestParameterParser.getFlag('a').value);
    ASSERT(validTestParameterParser.getFlag(std::string("aaaaa")).value);
    ASSERT(!validTestParameterParser.getFlag('b').isError);
    ASSERT(validTestParameterParser.getFlag('b').value);
    ASSERT(validTestParameterParser.getFlag(std::string("bbbbb")).value);
    ASSERT(!validTestParameterParser.getFlag('1').isError);
    ASSERT(validTestParameterParser.getFlag('1').value);
    ASSERT(validTestParameterParser.getFlag(std::string("11111")).value);
    ASSERT(!validTestParameterParser.getFlag('2').isError);
    ASSERT(validTestParameterParser.getFlag('2').value);
    ASSERT(validTestParameterParser.getFlag(std::string("22222")).value);
    ASSERT(!validTestParameterParser.getFlag('3').isError);
    ASSERT(validTestParameterParser.getFlag('3').value);
    ASSERT(validTestParameterParser.getFlag(std::string("33333")).value);
    ASSERT(!validTestParameterParser.getFlag('4').isError);
    ASSERT(validTestParameterParser.getFlag('4').value);
    ASSERT(validTestParameterParser.getFlag(std::string("44444")).value);
    ASSERT(!validTestParameterParser.getFlag('5').isError);
    ASSERT(validTestParameterParser.getFlag('5').value);
    ASSERT(validTestParameterParser.getFlag(std::string("55555")).value);
    ASSERT(!validTestParameterParser.getFlag(std::string("long-flag")).isError);
    ASSERT(validTestParameterParser.getFlag('l').value);
    ASSERT(validTestParameterParser.getFlag(std::string("long-flag")).value);
    ASSERT(!validTestParameterParser.getFlag('6').isError);
    ASSERT(!validTestParameterParser.getFlag('6').value);
    ASSERT(!validTestParameterParser.getFlag(std::string("66666")).value);
    ASSERT(!validTestParameterParser.getUnpositionalArgument('c').isError);
    ASSERT(validTestParameterParser.getUnpositionalArgument('c').value==std::string("concatenated"));
    ASSERT(validTestParameterParser.getUnpositionalArgument(std::string("ccccc")).value==std::string("concatenated"));
    ASSERT(!validTestParameterParser.getUnpositionalArgument('d').isError);
    ASSERT(validTestParameterParser.getUnpositionalArgument('d').value==std::string("separate-value"));
    ASSERT(validTestParameterParser.getUnpositionalArgument(std::string("ddddd")).value==std::string("separate-value"));
    ASSERT(!validTestParameterParser.getFlag('e').isError);
    ASSERT(validTestParameterParser.getFlag('e').value);
    ASSERT(validTestParameterParser.getFlag(std::string("eeeee")).value);
    ASSERT(!validTestParameterParser.getUnpositionalArgument('f').isError);
    ASSERT(validTestParameterParser.getUnpositionalArgument('f').value==std::string("double-concatenated"));
    ASSERT(validTestParameterParser.getUnpositionalArgument(std::string("fffff")).value==std::string("double-concatenated"));
    ASSERT(!validTestParameterParser.getPositionalArgument(0).isError);
    ASSERT(validTestParameterParser.getPositionalArgument(0).value=="positional argument 0");
    ASSERT(!validTestParameterParser.getFlag('g').isError);
    ASSERT(validTestParameterParser.getFlag('g').value);
    ASSERT(validTestParameterParser.getFlag(std::string("ggggg")).value);
    ASSERT(!validTestParameterParser.getUnpositionalArgument('h').isError);
    ASSERT(validTestParameterParser.getUnpositionalArgument('h').value==std::string("concatenated-separate-value"));
    ASSERT(validTestParameterParser.getUnpositionalArgument(std::string("hhhhh")).value==std::string("concatenated-separate-value"));
    ASSERT(!validTestParameterParser.getUnpositionalArgument('x').isError);
    ASSERT(validTestParameterParser.getUnpositionalArgument('x').value==std::string("included value"));
    ASSERT(validTestParameterParser.getUnpositionalArgument(std::string("long-argument-with-value-included")).value==std::string("included value"));
    ASSERT(!validTestParameterParser.getUnpositionalArgument('y').isError);
    ASSERT(validTestParameterParser.getUnpositionalArgument('y').value==std::string("separate value"));
    ASSERT(validTestParameterParser.getUnpositionalArgument(std::string("long-argument-with-value-separated")).value==std::string("separate value"));
    ASSERT(!validTestParameterParser.getPositionalArgument(1).isError);
    ASSERT(validTestParameterParser.getPositionalArgument(1).value=="positional argument 1");
    ASSERT(!validTestParameterParser.getPositionalArgument(2).isError);
    ASSERT(validTestParameterParser.getPositionalArgument(2).value=="positional argument 2");
    ASSERT(!validTestParameterParser.getUnpositionalArgument('z').isError);
    ASSERT(validTestParameterParser.getUnpositionalArgument('z').errorCode == ErrorCodes::NOT_PRESENT);
    ASSERT(validTestParameterParser.getUnpositionalArgument('z').value == std::string(""));

    delete[] validTestParameterList;
    std::cout << "Passed valid input test." << std::endl;

    // empty input (valid and invalid) #################################
    std::vector<CLI::Flag> emptyTestFlags;

    std::vector<CLI::UnpositionalArgument> emptyTestUnpositionalArguments;

    std::vector<CLI::PositionalArgument> emptyTestPositionalArguments;

    const char** emptyTestParameterList = new const char*[1];
    // The command is always a part of a command line, even if it is empty
    // otherwise. Therefore, I think it is reasonable to assume that it
    // will always be available.
    emptyTestParameterList[0] = "test";
    int emptyTestParameterCount = 1;

    // valid
    CLI::ArgumentsParser validEmptyTestParameterParser = CLI::ArgumentsParser(emptyTestParameterCount, emptyTestParameterList, emptyTestFlags, emptyTestUnpositionalArguments, emptyTestPositionalArguments);
    ASSERT(!validEmptyTestParameterParser.wrongUsage);
    ASSERT(validEmptyTestParameterParser.programName == std::string("test"));

    //invalid
    emptyTestPositionalArguments.push_back(CLI::PositionalArgument("argument", "positional argument"));

    CLI::ArgumentsParser invalidEmptyTestParameterParser = CLI::ArgumentsParser(emptyTestParameterCount, emptyTestParameterList, emptyTestFlags, emptyTestUnpositionalArguments, emptyTestPositionalArguments);
    ASSERT(invalidEmptyTestParameterParser.wrongUsage);
    ASSERT(invalidEmptyTestParameterParser.programName == std::string("test"));
    ASSERT(invalidEmptyTestParameterParser.getPositionalArgument(0).isError);
    ASSERT(invalidEmptyTestParameterParser.getPositionalArgument(0).errorCode == ErrorCodes::NOT_PRESENT);

    delete[] emptyTestParameterList;
    std::cout << "Passed empty input test." << std::endl;

    // unknown flag ####################################################

    std::vector<CLI::Flag> unknownFlagTestFlags;
    unknownFlagTestFlags.push_back(CLI::Flag('0', "zero", "test flag"));
    unknownFlagTestFlags.push_back(CLI::Flag('1', "one",  "test flag"));
    unknownFlagTestFlags.push_back(CLI::Flag('2', "two",  "test flag"));
    unknownFlagTestFlags.push_back(CLI::Flag('3', "three","test flag"));
    unknownFlagTestFlags.push_back(CLI::Flag('4', "four", "test flag"));
    unknownFlagTestFlags.push_back(CLI::Flag('5', "five", "test flag"));

    std::vector<CLI::UnpositionalArgument> unknownFlagTestUnpositionalArguments;

    std::vector<CLI::PositionalArgument> unknownFlagTestPositionalArguments;

    const char** unknownFlagTestParameterList = new const char*[7];
    unknownFlagTestParameterList[0] = "test";
    unknownFlagTestParameterList[1] = "-a";
    unknownFlagTestParameterList[2] = "-1";
    unknownFlagTestParameterList[3] = "-2";
    unknownFlagTestParameterList[4] = "-3";
    unknownFlagTestParameterList[5] = "-4";
    unknownFlagTestParameterList[6] = "-5";
    int unknownFlagStandaloneTestParameterCount = 2;

    CLI::ArgumentsParser unknownFlagStandaloneTestParser = CLI::ArgumentsParser(unknownFlagStandaloneTestParameterCount, unknownFlagTestParameterList, unknownFlagTestFlags, unknownFlagTestUnpositionalArguments, unknownFlagTestPositionalArguments);
    ASSERT(unknownFlagStandaloneTestParser.getFlag('1').isError);
    ASSERT(unknownFlagStandaloneTestParser.getFlag('1').errorCode == ErrorCodes::NOT_PRESENT);
    ASSERT(!unknownFlagStandaloneTestParser.getFlag('1').value);
    ASSERT(unknownFlagStandaloneTestParser.getFlag(std::string("four")).isError);
    ASSERT(unknownFlagStandaloneTestParser.getFlag(std::string("four")).errorCode == ErrorCodes::NOT_PRESENT);
    ASSERT(!unknownFlagStandaloneTestParser.getFlag(std::string("four")).value);

    int unknownFlagMixedTestParameterCount = 7;

    CLI::ArgumentsParser unknownFlagMixedTestParser = CLI::ArgumentsParser(unknownFlagMixedTestParameterCount, unknownFlagTestParameterList, unknownFlagTestFlags, unknownFlagTestUnpositionalArguments, unknownFlagTestPositionalArguments);
    ASSERT(unknownFlagMixedTestParser.getFlag('1').isError);
    ASSERT(unknownFlagMixedTestParser.getFlag('1').errorCode == ErrorCodes::WRONG_USAGE);
    ASSERT(unknownFlagMixedTestParser.getFlag('1').value);
    ASSERT(unknownFlagMixedTestParser.getFlag(std::string("four")).isError);
    ASSERT(unknownFlagMixedTestParser.getFlag(std::string("four")).errorCode == ErrorCodes::WRONG_USAGE);
    ASSERT(unknownFlagMixedTestParser.getFlag(std::string("four")).value);

    delete[] unknownFlagTestParameterList;
    std::cout << "Passed unknown flag test." << std::endl;

    // unknown unpositional argument ###################################

    std::vector<CLI::Flag> unknownArgumentTestFlags;

    std::vector<CLI::UnpositionalArgument> unknownArgumentTestUnpositionalArguments;
    unknownArgumentTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('0', "zero", "zero", "test flag"));
    unknownArgumentTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('1', "one",  "one",  "test flag"));
    unknownArgumentTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('2', "two",  "two",  "test flag"));
    unknownArgumentTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('3', "three","three","test flag"));
    unknownArgumentTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('4', "four", "four", "test flag"));
    unknownArgumentTestUnpositionalArguments.push_back(CLI::UnpositionalArgument('5', "five", "five", "test flag"));

    std::vector<CLI::PositionalArgument> unknownArgumentTestPositionalArguments;

    const char** unknownArgumentTestParameterList = new const char*[7];
    unknownArgumentTestParameterList[0] = "test";
    unknownArgumentTestParameterList[1] = "-a123";
    unknownArgumentTestParameterList[2] = "-1a";
    unknownArgumentTestParameterList[3] = "-2b";
    unknownArgumentTestParameterList[4] = "-3c";
    unknownArgumentTestParameterList[5] = "-4d";
    unknownArgumentTestParameterList[6] = "-5e";
    int unknownArgumentStandaloneTestParameterCount = 2;

    CLI::ArgumentsParser unknownArgumentStandaloneTestParser = CLI::ArgumentsParser(unknownArgumentStandaloneTestParameterCount, unknownArgumentTestParameterList, unknownArgumentTestFlags, unknownArgumentTestUnpositionalArguments, unknownArgumentTestPositionalArguments);
    ASSERT(unknownArgumentStandaloneTestParser.getUnpositionalArgument('1').isError);
    ASSERT(unknownArgumentStandaloneTestParser.getUnpositionalArgument('1').errorCode == ErrorCodes::NOT_PRESENT);
    ASSERT(unknownArgumentStandaloneTestParser.getUnpositionalArgument('1').value == "");
    ASSERT(unknownArgumentStandaloneTestParser.getUnpositionalArgument(std::string("four")).isError);
    ASSERT(unknownArgumentStandaloneTestParser.getUnpositionalArgument(std::string("four")).errorCode == ErrorCodes::NOT_PRESENT);
    ASSERT(unknownArgumentStandaloneTestParser.getUnpositionalArgument(std::string("four")).value == "");

    int unknownArgumentMixedTestParameterCount = 7;

    CLI::ArgumentsParser unknownArgumentMixedTestParser = CLI::ArgumentsParser(unknownArgumentMixedTestParameterCount, unknownArgumentTestParameterList, unknownArgumentTestFlags, unknownArgumentTestUnpositionalArguments, unknownArgumentTestPositionalArguments);
    ASSERT(unknownArgumentMixedTestParser.getUnpositionalArgument('1').isError);
    ASSERT(unknownArgumentMixedTestParser.getUnpositionalArgument('1').errorCode == ErrorCodes::WRONG_USAGE);
    ASSERT(unknownArgumentMixedTestParser.getUnpositionalArgument('1').value == "a");
    ASSERT(unknownArgumentMixedTestParser.getUnpositionalArgument(std::string("four")).isError);
    ASSERT(unknownArgumentMixedTestParser.getUnpositionalArgument(std::string("four")).errorCode == ErrorCodes::WRONG_USAGE);
    ASSERT(unknownArgumentMixedTestParser.getUnpositionalArgument(std::string("four")).value == "d");

    delete[] unknownArgumentTestParameterList;
    std::cout << "Passed unknown unpositional argument test." << std::endl;

    return 0;
}
